#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#


import os
import sys
import unittest
import warnings
import sasoptpy as so
from inspect import cleandoc
from sasoptpy.actions import read_data

from tests.swat_config import create_cas_connection


class TestReadData(unittest.TestCase):
    """
    Unit tests for READ DATA statements
    """

    @classmethod
    def setUpClass(cls):
        so.reset()
        cls.conn = None
        from swat import CAS, SWATError
        try:
            cls.conn = create_cas_connection()
        except SWATError:
            warnings.warn('CAS connection is not available', RuntimeWarning)
        except TypeError:
            warnings.warn('CAS variables are not available', RuntimeWarning)

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.close()

    def setUp(self):
        pass

    def test_read_regular_existing(self):

        from sasoptpy import Workspace, VariableGroup
        from sasoptpy.abstract import Set, ParameterGroup
        from sasoptpy.actions import solve

        with Workspace('test_workspace') as ws:

            ITEMS = Set(name='ITEMS')
            value = ParameterGroup(ITEMS, name='value', init=0)
            get = VariableGroup(ITEMS, name='get', vartype=so.INT, lb=0)

            read_data(
                table="values",
                index={'target': ITEMS, 'key': None},
                columns=[{'target': value}])

            solve(options={'with': so.BLACKBOX, 'maxgen': 10})

        optmodel_code = so.to_optmodel(ws)
        self.assertEqual(optmodel_code, cleandoc("""
            proc optmodel;
               set ITEMS;
               num value {ITEMS} init 0;
               var get {{ITEMS}} integer >= 0;
               read data values into ITEMS value;
               solve with blackbox / maxgen=10;
            quit;
            """))

    def test_read_data_index(self):

        with so.Workspace('test_read_data_n') as ws:
            ASSETS = so.Set(name='ASSETS')
            ret = so.ParameterGroup(ASSETS, name='return', ptype=so.NUM)
            read_data(
                table='means',
                index={'target': ASSETS, 'key': so.N},
                columns=[{'target': ret}]
            )
        optmodel_code = so.to_optmodel(ws)
        self.assertEqual(optmodel_code, cleandoc("""
            proc optmodel;
               set ASSETS;
               num return {ASSETS};
               read data means into ASSETS=[_N_] return;
            quit;
            """))

    def test_read_data_no_index_expression(self):

        from sasoptpy.util import iterate

        with so.Workspace('test_read_data_no_index_expression') as ws:
            ASSETS = so.Set(name='ASSETS')
            cov = so.ParameterGroup(ASSETS, ASSETS, name='cov', init=0)
            with iterate(ASSETS, 'asset1') as asset1, iterate(ASSETS, 'asset2') as asset2:
                read_data(
                    table='covdata',
                    index={'key': [asset1, asset2]},
                    columns=[
                        {'target': cov},
                        {'target': cov[asset2, asset1],
                         'column': 'cov'}
                    ]
                )

        optmodel_code = so.to_optmodel(ws)
        self.assertEqual(optmodel_code, cleandoc("""
            proc optmodel;
               set ASSETS;
               num cov {ASSETS, ASSETS} init 0;
               read data covdata into [asset1 asset2] cov cov[asset2, asset1]=cov;
            quit;
            """))

    def test_read_data_col(self):

        from sasoptpy.util import concat

        with so.Workspace('test_read_data_col') as ws:
            n = so.Parameter(name='n', init=2)
            indx = so.Set(name='indx', settype=['num'])
            p = so.ParameterGroup(indx, name='p')
            q = so.ParameterGroup(indx, name='q')

            read_data(
                table='exdata',
                index={
                    'target': indx,
                    'key': so.N
                },
                columns=[
                    {'target': p, 'column': 'column1'},
                    {'target': q, 'column': concat('column', n)}
                ]
            )

        optmodel_code = so.to_optmodel(ws)
        self.assertEqual(optmodel_code, cleandoc("""
            proc optmodel;
               num n init 2;
               set indx;
               num p {indx};
               num q {indx};
               read data exdata into indx=[_N_] p=column1 q=col('column' || n);
            quit;
            """))

    def test_read_data_indexed_column(self):

        from sasoptpy.util import iterate, concat

        with so.Workspace(name='test_read_data_idx_col') as ws:
            dow = so.Set(name='DOW', value=so.exp_range(1, 6))
            locs = so.Set(name='LOCS', settype=so.STR)
            demand = so.ParameterGroup(locs, dow, name='demand')

            with iterate(locs, name='loc') as loc:
                r = read_data(
                    table='dmnd',
                    index={'target': locs, 'key': loc}
                )
                with iterate(dow, name='d') as d:
                    r.append({
                        'index': d,
                        'target': demand[loc, d],
                        'column': concat('day', d)
                    })

        optmodel_code = so.to_optmodel(ws)
        self.assertEqual(optmodel_code, cleandoc("""
            proc optmodel;
               set DOW = 1..5;
               set <str> LOCS;
               num demand {LOCS, DOW};
               read data dmnd into LOCS=[loc] {d in DOW} < demand[loc, d]=col('day' || d) >;
            quit;"""))

    def test_with_cas_data(self):
        if TestReadData.conn is None:
            self.skipTest('No session is available')

        from sasoptpy.util import concat
        import pandas as pd
        df = pd.DataFrame([
            [1, 2], [3, 4]
        ], columns=['column1', 'column2'])

        s = TestReadData.conn
        exdata = s.upload_frame(df, casout='exdata')

        with so.Workspace('test_read_data_col', session=s) as ws:
            n = so.Parameter(name='n', init=2)
            indx = so.Set(name='indx', settype=['num'])
            p = so.ParameterGroup(indx, name='p')
            q = so.ParameterGroup(indx, name='q')

            read_data(
                table=exdata,
                index={
                    'target': indx,
                    'key': so.N
                },
                columns=[
                    {'target': p, 'column': 'column1'},
                    {'target': q, 'column': concat('column', n)}
                ]
            )

            from sasoptpy.actions import print_item
            print_item(p, q)

        optmodel_code = so.to_optmodel(ws)
        self.assertEqual(optmodel_code, cleandoc("""
            proc optmodel;
               num n init 2;
               set indx;
               num p {indx};
               num q {indx};
               read data EXDATA into indx=[_N_] p=column1 q=col('column' || n);
               print p q;
            quit;
            """))

        ws.submit(verbose=True)
        self.assertEqual(ws.response['Print1.PrintTable'].to_string(), cleandoc(
            """
            Print Table
            
               COL1    p    q
            0   1.0  1.0  2.0
            1   2.0  3.0  4.0
            """
        ))

    def test_read_data_no_index_target(self):

        with so.Workspace('test_no_index_target') as ws:
            NODES = so.Set(name='NODES', settype=so.STR)
            nodesByPri = so.ParameterGroup(NODES, name='nodesByPri')

            read_data(table='temppri',
                      index={'key': so.N},
                      columns=[
                          {'target': nodesByPri, 'column': 'id'}
                      ])

        self.assertEqual(so.to_optmodel(ws), cleandoc("""
            proc optmodel;
               set <str> NODES;
               num nodesByPri {NODES};
               read data temppri into [_N_] nodesByPri=id;
            quit;
            """))

    def test_read_data_same_name_cols(self):

        from sasoptpy.actions import set_objective, solve

        with so.Workspace('test_same_name_cols') as ws:
            x = so.Variable(name='x')
            y = so.Variable(name='y')
            c_xx = so.Parameter(name='c_xx')
            c_x = so.Parameter(name='c_x')
            c_y = so.Parameter(name='c_y')
            c_xy = so.Parameter(name='c_xy')
            c_yy = so.Parameter(name='c_yy')
            read_data(table='coeff',
                      index={},
                      columns=[
                          {'target': c_xx, 'column': 'c_xx'},
                          {'target': c_x},
                          {'column': 'c_y'},
                          {'target': c_xy, 'column': 'c_xy'},
                          {'target': c_yy}
                      ])
            set_objective(c_xx * x**2 + c_x * x + c_xy * x * y + c_yy * y**2,
                          name='z', sense=so.MIN)
            solve()

        self.assertEqual(so.to_optmodel(ws), cleandoc("""
            proc optmodel;
               var x;
               var y;
               num c_xx;
               num c_x;
               num c_y;
               num c_xy;
               num c_yy;
               read data coeff into c_xx c_x c_y c_xy c_yy;
               MIN z = c_xx * (x) ^ (2) + c_x * x + c_xy * x * y + c_yy * (y) ^ (2);
               solve;
            quit;
            """))

    def test_read_data_var_bound(self):

        from sasoptpy.util import iterate

        with so.Workspace('test_read_data_var_bound') as ws:
            plants = so.Set(name='plants', settype=so.STR)
            prod = so.VariableGroup(plants, name='prod', lb=0)
            cost = so.Parameter(name='cost', ptype=so.NUM)
            p = so.SetIterator(plants, name='p')
            r = read_data(
                table='pdat',
                index={'target': plants, 'key': p},
                columns=[
                    {'target': prod[p].ub, 'column': 'maxprod'},
                    {'target': cost}
                ])

        self.assertEqual(so.to_optmodel(ws), cleandoc("""
            proc optmodel;
               set <str> plants;
               var prod {{plants}} >= 0;
               num cost;
               read data pdat into plants=[p] prod[p].ub=maxprod cost;
            quit;
            """))

    def test_read_data_with_exp(self):

        with so.Workspace('test_read_data_with_exp') as ws:

            ss = so.Set(name='subscripts', value=so.exp_range(1, 5), settype=so.NUM)
            letter = so.ParameterGroup(ss, name='letter', ptype=so.STR)
            read_data(
                table='abcd',
                index={'key': so.N},
                columns=[
                    {'target': letter[5-so.N]}
                ]
            )

        self.assertEqual(so.to_optmodel(ws), cleandoc("""
            proc optmodel;
               set subscripts = 1..4;
               str letter {subscripts};
               read data abcd into [_N_] letter[- _N_ + 5];
            quit;
            """))

    def test_read_data_append_later(self):

        from sasoptpy.util import concat, iterate

        with so.Workspace('let') as ws:
            actors = so.Set(name='ACTORS')
            actor_name = so.ParameterGroup(actors, ptype=so.STR,
                                           name='actor_name')
            daily_fee = so.ParameterGroup(actors, name='daily_fee')
            most_scenes = so.Parameter(name='most_scenes', value=9)
            scene_list = so.ParameterGroup(actors, so.exp_range(1, most_scenes),
                                           name='scene_list')

            r = read_data(
                table='scene',
                index={
                    'target': actors,
                    'key': so.N
                },
                columns=[
                    {
                        'target': actor_name,
                        'column': 'Actor'
                    },
                    {
                        'target': daily_fee,
                        'column': 'DailyFee'
                    }
                ]
            )
            with iterate(so.exp_range(1, most_scenes), name='j') as j:
                r.append({
                    'index': j,
                    'target': scene_list[so.N, j],
                    'column': concat('S_Var', j)
                })

        self.assertEqual(so.to_optmodel(ws), cleandoc("""
            proc optmodel;
               set ACTORS;
               str actor_name {ACTORS};
               num daily_fee {ACTORS};
               num most_scenes = 9;
               num scene_list {ACTORS, 1..most_scenes};
               read data scene into ACTORS=[_N_] actor_name=Actor daily_fee=DailyFee {j in 1..most_scenes} < scene_list[_N_, j]=col('S_Var' || j) >;
            quit;
            """))

    def test_read_data_multi_index(self):

        with so.Workspace('test_read_data_multi_index') as ws:
            arcs = so.Set(name='ARCS', settype=[so.STR, so.STR])
            arc_lower = so.ParameterGroup(arcs, name='arcLower')
            arc_upper = so.ParameterGroup(arcs, name='arcUpper')
            arc_cost = so.ParameterGroup(arcs, name='arcCost')

            read_data(
                table='arcdata',
                index={
                    'target': arcs,
                    'key': ['_tail_', '_head_']
                },
                columns=[
                    {'target': arc_lower, 'column': '_lo_'},
                    {'target': arc_upper, 'column': '_capac_'},
                    {'target': arc_cost, 'column': '_cost_'}
                ]
            )

        self.assertEqual(so.to_optmodel(ws), cleandoc("""
            proc optmodel;
               set <str, str> ARCS;
               num arcLower {ARCS};
               num arcUpper {ARCS};
               num arcCost {ARCS};
               read data arcdata into ARCS=[_tail_ _head_] arcLower=_lo_ arcUpper=_capac_ arcCost=_cost_;
            quit;
            """))

    def test_read_data_single_col(self):

        from sasoptpy.util import concat, iterate

        with so.Workspace('test_read_data_single_col') as ws:
            daily_employee_slots = so.Set(
                name='DailyEmployeeSlots', settype=[so.STR, so.NUM])
            weekdays = so.Set(
                name='WeekDays', value=['mon', 'tue', 'wed', 'thu', 'fri'])
            preference_weights = so.ParameterGroup(
                daily_employee_slots, weekdays, name='PreferenceWeights')

            with iterate(daily_employee_slots, name=['name', 'slot']) as keys:
                r = read_data(
                    table='preferences',
                    index={
                        'target': daily_employee_slots,
                        'key': keys
                    }
                )
                with iterate(weekdays, name='day') as day:
                    r.append({
                        'index': day,
                        'target': preference_weights[keys['name'], keys['slot'], day],
                        'column': day
                    })

        optmodel_code = so.to_optmodel(ws)
        self.assertEqual(optmodel_code, cleandoc("""
            proc optmodel;
               set <str, num> DailyEmployeeSlots;
               set WeekDays = {'mon','tue','wed','thu','fri'};
               num PreferenceWeights {DailyEmployeeSlots, WeekDays};
               read data preferences into DailyEmployeeSlots=[name slot] {day in WeekDays} < PreferenceWeights[name, slot, day]=col(day) >;
            quit;
            """))

    def test_read_data_N_in_index(self):

        from sasoptpy.util import concat

        with so.Workspace('read_data_N_as_index') as ws:
            tasks = so.Set(name='TASKS', value=so.exp_range(1, 25))
            machines = so.Set(name='MACHINES', value=so.exp_range(1, 9))
            profit = so.ParameterGroup(machines, tasks, name='profit')
            j = so.SetIterator(tasks, name='j')
            read_data(
                table='profit_data',
                index={'key': so.N},
                columns=[
                    {'index': j, 'target': profit[so.N, j], 'column': concat('p', j)}
                ]
            )

        self.assertEqual(so.to_optmodel(ws), cleandoc("""
            proc optmodel;
               set TASKS = 1..24;
               set MACHINES = 1..8;
               num profit {MACHINES, TASKS};
               read data profit_data into [_N_] {j in TASKS} < profit[_N_, j]=col('p' || j) >;
            quit;
            """))

    def test_with_model(self):

        m = so.Model(name='m')
        ITEMS = m.add_set(name='ITEMS')
        value = m.add_parameter(ITEMS, name='value', init=0)
        weight = m.add_parameter(ITEMS, name='weight')
        limit = m.add_parameter(ITEMS, name='limit')
        get = m.add_variables(ITEMS, name='get', vartype=so.INT, lb=0)
        m.set_objective(so.expr_sum(get[i] for i in ITEMS), name='max_get', sense=so.MAX)

        m.include(read_data(table='values', index={'target': ITEMS, 'key': None},
                            columns=[value, weight, limit]))

        self.assertEqual(
            so.to_optmodel(m, options={'with': so.BLACKBOX, 'maxgen': 10}),
            cleandoc('''
            proc optmodel;
               set ITEMS;
               num value {{ITEMS}} init 0;
               num weight {{ITEMS}};
               num limit {{ITEMS}};
               var get {{ITEMS}} integer >= 0;
               max max_get = sum {i in ITEMS} (get[i]);
               read data values into ITEMS value weight limit;
               solve with blackbox / maxgen=10;
            quit;'''))


    def tearDown(self):
        pass



import sasoptpy as so
import pandas as pd


def test(cas_conn):

    m = so.Model(name='decentralization', session=cas_conn)

    DEPTS = ['A', 'B', 'C', 'D', 'E']
    CITIES = ['Bristol', 'Brighton', 'London']

    benefit_data = pd.DataFrame([
        ['Bristol', 10, 15, 10, 20, 5],
        ['Brighton', 10, 20, 15, 15, 15]],
        columns=['city'] + DEPTS).set_index('city')

    comm_data = pd.DataFrame([
        ['A', 'B', 0.0],
        ['A', 'C', 1.0],
        ['A', 'D', 1.5],
        ['A', 'E', 0.0],
        ['B', 'C', 1.4],
        ['B', 'D', 1.2],
        ['B', 'E', 0.0],
        ['C', 'D', 0.0],
        ['C', 'E', 2.0],
        ['D', 'E', 0.7]], columns=['i', 'j', 'comm']).set_index(['i', 'j'])

    cost_data = pd.DataFrame([
        ['Bristol', 'Bristol', 5],
        ['Bristol', 'Brighton', 14],
        ['Bristol', 'London', 13],
        ['Brighton', 'Brighton', 5],
        ['Brighton', 'London', 9],
        ['London', 'London', 10]], columns=['i', 'j', 'cost']).set_index(
            ['i', 'j'])

    max_num_depts = 3

    benefit = {}
    for city in CITIES:
        for dept in DEPTS:
            try:
                benefit[dept, city] = benefit_data.loc[city, dept]
            except:
                benefit[dept, city] = 0

    comm = {}
    for row in comm_data.iterrows():
        (i, j) = row[0]
        comm[i, j] = row[1]['comm']
        comm[j, i] = comm[i, j]

    cost = {}
    for row in cost_data.iterrows():
        (i, j) = row[0]
        cost[i, j] = row[1]['cost']
        cost[j, i] = cost[i, j]

    assign = m.add_variables(DEPTS, CITIES, vartype=so.BIN, name='assign')
    IJKL = [(i, j, k, l)
            for i in DEPTS for j in CITIES for k in DEPTS for l in CITIES
            if i < k]
    product = m.add_variables(IJKL, vartype=so.BIN, name='product')

    totalBenefit = so.expr_sum(benefit[i, j] * assign[i, j]
                                for i in DEPTS for j in CITIES)

    totalCost = so.expr_sum(comm[i, k] * cost[j, l] * product[i, j, k, l]
                             for (i, j, k, l) in IJKL)

    m.set_objective(totalBenefit-totalCost, name='netBenefit', sense=so.MAX)

    m.add_constraints((so.expr_sum(assign[dept, city] for city in CITIES)
                      == 1 for dept in DEPTS), name='assign_dept')

    m.add_constraints((so.expr_sum(assign[dept, city] for dept in DEPTS)
                      <= max_num_depts for city in CITIES), name='cardinality')

    product_def1 = m.add_constraints((assign[i, j] + assign[k, l] - 1
                                     <= product[i, j, k, l]
                                     for (i, j, k, l) in IJKL),
                                     name='pd1')

    product_def2 = m.add_constraints((product[i, j, k, l] <= assign[i, j]
                                      for (i, j, k, l) in IJKL),
                                     name='pd2')

    product_def3 = m.add_constraints((product[i, j, k, l] <= assign[k, l]
                                      for (i, j, k, l) in IJKL),
                                     name='pd3')

    m.solve()
    print(m.get_problem_summary())

    m.drop_constraints(product_def1)
    m.drop_constraints(product_def2)
    m.drop_constraints(product_def3)

    m.add_constraints((
        so.expr_sum(product[i, j, k, l]
                     for j in CITIES if (i, j, k, l) in IJKL) == assign[k, l]
        for i in DEPTS for k in DEPTS for l in CITIES if i < k),
        name='pd4')

    m.add_constraints((
        so.expr_sum(product[i, j, k, l]
                     for l in CITIES if (i, j, k, l) in IJKL) == assign[i, j]
        for k in DEPTS for i in DEPTS for j in CITIES if i < k),
        name='pd5')

    m.solve()
    print(m.get_problem_summary())
    totalBenefit.set_name('totalBenefit')
    totalCost.set_name('totalCost')
    print(so.get_solution_table(totalBenefit, totalCost))
    print(so.get_solution_table(assign).unstack(level=-1))

    return m.get_objective_value()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import unittest
from inspect import cleandoc

import sasoptpy as so


class TestPackageUtils(unittest.TestCase):

    def setUp(self):
        so.reset()

    def test_extract_arguments(self):
        x = so.VariableGroup(['a', 'b', 'c'], [1, 2, 3], name='x')
        def1 = so.to_definition(x)

        y = so.VariableGroup(('a', 'b', 'c'), (1, 2, 3), name='x')
        def2 = so.to_definition(y)

        self.assertEqual(def1, def2)

    def test_extract_list_value(self):
        m = so.Model(name='test_extract_list_vals')
        S = ['a', 'b', 'c']
        lb_values = {'a': 1, 'b': 0, 'c': 2}
        ub_values = {'a': 5, 'b': 10}
        init_values = {'b': 2, 'c': 3}
        x = m.add_variables(S, name='x', ub=ub_values, lb=lb_values,
                             init=init_values)
        self.assertEqual(so.to_optmodel(m), cleandoc('''
            proc optmodel;
               min test_extract_list_vals_obj = 0;
               var x {{'a','b','c'}};
               x['a'].lb = 1;
               x['a'].ub = 5;
               x['b'] = 2;
               x['b'].lb = 0;
               x['b'].ub = 10;
               x['c'] = 3;
               x['c'].lb = 2;
               solve;
            quit;
            '''))

        def produce_error():
            from collections import OrderedDict
            ind = ['a', 'b', 'c']
            y_lb = set([0, 1, 2])
            y = m.add_variables(ind, name='y', lb=y_lb)
        self.assertRaises(ValueError, produce_error)

    def test_deprecation(self):
        def call_tuple_unpack():
            so.util.tuple_unpack((1,2,))
        self.assertWarns(DeprecationWarning, call_tuple_unpack)
        def call_tuple_pack():
            so.util.tuple_pack(1)
        self.assertWarns(DeprecationWarning, call_tuple_pack)
        def call_list_pack():
            so.util.list_pack((1,2,3))
        self.assertWarns(DeprecationWarning, call_list_pack)
        def call_wrap():
            so.util.wrap(5)
        self.assertWarns(DeprecationWarning, call_wrap)

    def test_sum_wrap(self):
        x = so.Variable(name='x')
        e = so.expr_sum(x for _ in range(3))
        self.assertEqual(so.to_expression(e), '3 * x')

    def test_sum_wrap_abstract(self):
        I = so.Set(name='I')
        x = so.Variable(name='x')
        e = so.expr_sum(x for i in I)
        self.assertEqual(so.to_expression(e), 'sum {i in I} (x)')

    def test_comparable(self):
        self.assertTrue(so.util.is_comparable(4))
        self.assertFalse(so.util.is_comparable(dict()))
        self.assertTrue(so.util.is_comparable('abc'))

    def test_flatten_tuple(self):
        tp = (3, 4, (5, (1, 0), 2))
        self.assertEqual(list(so.util.flatten_tuple(tp)),
                         [3, 4, 5, 1, 0, 2])

    def test_sas_string(self):
        S = so.exp_range(1, 11, 2)
        self.assertEqual(so.util.package_utils._to_sas_string(S), '1..10 by 2')

        def invalid_type():
            from collections import OrderedDict
            so.util.package_utils._to_sas_string(OrderedDict(a=4))
        self.assertRaises(TypeError, invalid_type)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for core classes.
"""

from difflib import SequenceMatcher
import inspect
import unittest

import sasoptpy as so
from sasoptpy.core.util import *


def match_ratio(a, b):
    seq = SequenceMatcher(None, a, b)
    return seq.ratio()


class TestUtil(unittest.TestCase):
    """
    Unit tests for core utility functions
    """

    @classmethod
    def setUpClass(cls):
        cls.x = x = so.Variable(name='x')
        cls.e = so.Expression(2 * x, name='e')
        cls.c = so.Constraint(4 * x**2 <= 10, name='c')
        cls.I = I = so.abstract.Set(name='I')
        cls.a = so.VariableGroup(I, name='a')
        cls.S1 = S1 = so.abstract.Set(name='S1')
        cls.S2 = S2 = so.abstract.Set(name='S2', settype='str')
        cls.S3 = S3 = so.abstract.Set(name='S3')
        cls.p = so.abstract.ParameterGroup(S1, name='p')
        cls.y = so.VariableGroup(S1, S2, S3, name='y')

    def test_expression(self):
        self.assertTrue(is_expression(TestUtil.e))
        self.assertTrue(is_expression(TestUtil.x))
        self.assertTrue(is_expression(TestUtil.c))

    def test_variable(self):
        self.assertFalse(is_variable(TestUtil.e))
        self.assertTrue(is_variable(TestUtil.x))
        self.assertFalse(is_variable(TestUtil.c))

    def test_constraint(self):
        self.assertTrue(is_constraint(TestUtil.c))
        self.assertFalse(is_constraint(TestUtil.e))

    def test_abstract(self):
        self.assertTrue(is_abstract(TestUtil.a))
        self.assertFalse(is_abstract(TestUtil.x))

    def test_itearor_expression(self):
        so.reset()
        S1 = TestUtil.S1
        S2 = TestUtil.S2
        S3 = TestUtil.S3
        p = TestUtil.p
        y = TestUtil.y
        c = so.ConstraintGroup((y[i, 'a', 1] + p[i] <= 5 for i in S1),
                               name='it_exp')
        self.assertEqual(so.to_definition(c),
                         "con it_exp {{o{i} in S1}} : "
                         "y[o{i}, 'a', 1] + p[o{i}] <= 5;".format(i=1))

    def test_safe_iterator_expression(self):
        S1 = TestUtil.S1
        S2 = ['a', 'b c', 'd']
        S3 = [1, 2, 3]
        y = TestUtil.y
        c = so.ConstraintGroup((y [i, j, k] <= 5 for i in S1 for j in S2 for k in S3),
                               name='safe_it_exp')
        self.assertEqual(so.to_definition(c), inspect.cleandoc(
            """
            con safe_it_exp_a_1 {o1 in S1} : y[o1, 'a', 1] <= 5;
            con safe_it_exp_a_2 {o1 in S1} : y[o1, 'a', 2] <= 5;
            con safe_it_exp_a_3 {o1 in S1} : y[o1, 'a', 3] <= 5;
            con safe_it_exp_b_c_1 {o1 in S1} : y[o1, 'b c', 1] <= 5;
            con safe_it_exp_b_c_2 {o1 in S1} : y[o1, 'b c', 2] <= 5;
            con safe_it_exp_b_c_3 {o1 in S1} : y[o1, 'b c', 3] <= 5;
            con safe_it_exp_d_1 {o1 in S1} : y[o1, 'd', 1] <= 5;
            con safe_it_exp_d_2 {o1 in S1} : y[o1, 'd', 2] <= 5;
            con safe_it_exp_d_3 {o1 in S1} : y[o1, 'd', 3] <= 5;
            """
        ))

    def test_evaluate(self):
        x = TestUtil.x
        import sasoptpy.abstract.math as sm
        import math

        e1 = x * x + x - 5
        e2 = 2 * x ** 2 - 10
        e3 = 2 / x + 5 + sm.sin(x)

        x.set_value(3)
        self.assertEqual(e1.get_value(), 7)
        self.assertEqual(e2.get_value(), 8)
        self.assertEqual(e3.get_value(), 17/3 + math.sin(3))
        x.set_value(4)
        self.assertEqual(e1.get_value(), 15)
        self.assertEqual(e2.get_value(), 22)
        self.assertEqual(e3.get_value(), 5.5 + math.sin(4))

        def division_by_zero():
            x.set_value(0)
            e3.get_value()
        self.assertRaises(ZeroDivisionError, division_by_zero)

        x.set_value(2)
        e4 = x ** (sm.sin(x)) - 1
        self.assertEqual(e4.get_value(), 2 ** math.sin(2) - 1)

    def test_expression_to_constraint(self):
        x = TestUtil.x
        e1 = 4 * x - 10

        c = so.core.util.expression_to_constraint(e1, 'E', [2, 10])
        self.assertTrue(so.to_definition(c), "con None : 12 <= 4 * x <= 20;")

        c = so.core.util.expression_to_constraint(x, 'G', 10)
        self.assertEqual(so.to_definition(c), "con None : x >= 10;")

        c = so.core.util.expression_to_constraint(2 * x, 'L', 3 * x)
        self.assertEqual(so.to_definition(c), "con None : - x <= 0;")

        c = so.core.util.expression_to_constraint(x, 'G', 0)
        self.assertEqual(so.to_definition(c), "con None : x >= 0;")

        import sasoptpy.abstract.math as sm
        c = so.core.util.expression_to_constraint(sm.sin(x), 'L', 1)
        self.assertEqual(so.to_definition(c), "con None : sin(x) <= 1;")

        y = TestUtil.y
        S1 = TestUtil.S1
        e1 = so.expr_sum(y[i, 'a', 1] for i in S1)
        c = so.core.util.expression_to_constraint(e1, 'G', 5)
        self.assertEqual(so.to_definition(c), "con None : sum {i in S1} "
                                              "(y[i, 'a', 1]) >= 5;")

    def tearDown(self):
        so.reset()

    @classmethod
    def tearDownClass(self):
        so.reset()





#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Test for set class
"""

import unittest
import sasoptpy as so
from inspect import cleandoc

import os
import sys
current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '..')))
from util import assert_equal_wo_temps

class TestSet(unittest.TestCase):
    """
    Unit tests for :class:`sasoptpy.abstract.Set` objects
    """

    def setUp(self):
        so.reset()

    def test_initialization(self):
        with so.Workspace('w') as w:
            S = so.Set(name='S')
            T = so.Set(name='T', init=range(11))
            V = so.Set(name='V', value=[1, 2, 4])
            W = so.Set(name='W', settype=[so.STR, so.NUM])

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set S;
               set T init 0..10;
               set V = {1,2,4};
               set <str, num> W;
            quit;'''))

        self.assertEqual(repr(S),
                         'sasoptpy.abstract.Set(name=S, settype=[\'num\'])')
        self.assertEqual(
            repr(T),
            'sasoptpy.abstract.Set(name=T, settype=[\'num\'], init=range(0, 11))')
        self.assertEqual(
            repr(V),
            'sasoptpy.abstract.Set(name=V, settype=[\'num\'], value=[1, 2, 4])'
        )
        self.assertEqual(
            repr(W),
            'sasoptpy.abstract.Set(name=W, settype=[\'str\', \'num\'])'
        )

    def test_in_set(self):

        e = so.Parameter(name='e', value=1)
        S = so.Set(name='S')
        if e.sym in S:
            self.assertEqual(e.sym.get_conditions_str(), 'e IN S')
        x = so.Variable(name='x')
        def incorrect_type():
            if x in S:
                print('x is in S')
        self.assertRaises(RuntimeError, incorrect_type)

    def test_inline_set(self):
        from sasoptpy.actions import inline_condition, for_loop
        from sasoptpy.abstract.math import mod

        with so.Workspace('w') as w:
            p = so.Parameter(name='p')
            S = so.Set(name='S', value=so.exp_range(1, 11))
            iset = so.InlineSet(lambda: (x for x in S
                                if inline_condition(mod(x, 3) == 0)))
            for i in for_loop(iset):
                p.set_value(i)

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc('''
            proc optmodel;
               num p;
               set S = 1..10;
               for {TEMP2 in {TEMP1 in S: mod(TEMP1 , 3) = 0}} do;
                  p = TEMP2;
               end;
            quit;'''))

        assert_equal_wo_temps(self, repr(iset),
                              'sasoptpy.InlineSet({o4 in S: mod(o4 , 3) = 0}}')


#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from inspect import cleandoc
import os
import unittest
import warnings

from swat import CAS, SWATError

import sasoptpy as so

from tests.swat_config import create_cas_connection


class TestOPTMODEL(unittest.TestCase):
    """
    Unit tests for the CAS interface
    """

    @classmethod
    def setUpClass(cls):
        so.reset()
        cls.conn = None
        try:
            cls.conn = create_cas_connection()
        except SWATError:
            warnings.warn('CAS connection is not available', RuntimeWarning)
        except TypeError:
            warnings.warn('CAS variables are not available', RuntimeWarning)

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.close()

    def test_variable_group_assignments(self):

        from sasoptpy.actions import read_data

        if TestOPTMODEL.conn is None:
            self.skipTest('CAS Session is not available')

        import pandas as pd
        df = pd.DataFrame([
            ['a', 'b', 1],
            ['c', 'd,e', 2],
            ['f,g', 'g,h,i', 3]
        ], columns=['k1', 'k2', 'v'])


        m = so.Model(name='m', session=TestOPTMODEL.conn)
        setK1 = df['k1'].tolist()
        setK2 = df['k2'].tolist()

        x = m.add_variables(setK1, setK2, name='x')
        m.add_constraints((x[i, j] >= 1 for i in setK1 for j in setK2), name='c')
        m.set_objective(so.expr_sum(x[i, j] for i in setK1 for j in setK2), name='obj', sense=so.minimize)
        m.solve(verbose=True)
        self.assertEqual(str(m.get_solution()), cleandoc('''
            Selected Rows from Table SOLUTION

                 i               var  value             lb             ub   rc
            0  1.0            x[a,b]    1.0 -1.797693e+308  1.797693e+308  0.0
            1  2.0        x[a,'d,e']    1.0 -1.797693e+308  1.797693e+308  0.0
            2  3.0      x[a,'g,h,i']    1.0 -1.797693e+308  1.797693e+308  0.0
            3  4.0            x[c,b]    1.0 -1.797693e+308  1.797693e+308  0.0
            4  5.0        x[c,'d,e']    1.0 -1.797693e+308  1.797693e+308  0.0
            5  6.0      x[c,'g,h,i']    1.0 -1.797693e+308  1.797693e+308  0.0
            6  7.0        x['f,g',b]    1.0 -1.797693e+308  1.797693e+308  0.0
            7  8.0    x['f,g','d,e']    1.0 -1.797693e+308  1.797693e+308  0.0
            8  9.0  x['f,g','g,h,i']    1.0 -1.797693e+308  1.797693e+308  0.0'''))

        so.config['generic_naming'] = True
        self.assertEqual(m.to_optmodel(), cleandoc('''
            proc optmodel;
               var x {{'a','c','f,g'}, {'b','d,e','g,h,i'}};
               x['a', 'b'] = 1.0;
               x['a', 'd,e'] = 1.0;
               x['a', 'g,h,i'] = 1.0;
               x['c', 'b'] = 1.0;
               x['c', 'd,e'] = 1.0;
               x['c', 'g,h,i'] = 1.0;
               x['f,g', 'b'] = 1.0;
               x['f,g', 'd,e'] = 1.0;
               x['f,g', 'g,h,i'] = 1.0;
               con c_0 : x['a', 'b'] >= 1;
               con c_1 : x['a', 'd,e'] >= 1;
               con c_2 : x['a', 'g,h,i'] >= 1;
               con c_3 : x['c', 'b'] >= 1;
               con c_4 : x['c', 'd,e'] >= 1;
               con c_5 : x['c', 'g,h,i'] >= 1;
               con c_6 : x['f,g', 'b'] >= 1;
               con c_7 : x['f,g', 'd,e'] >= 1;
               con c_8 : x['f,g', 'g,h,i'] >= 1;
               min obj = x['a', 'b'] + x['a', 'd,e'] + x['a', 'g,h,i'] + x['c', 'b'] + x['c', 'd,e'] + x['c', 'g,h,i'] + x['f,g', 'b'] + x['f,g', 'd,e'] + x['f,g', 'g,h,i'];
               solve;
            quit;'''))




import sasoptpy as so


def test(cas_conn, sols=False):

    m = so.Model(name='multiobjective', session=cas_conn)

    x = m.add_variables([1, 2], lb=0, ub=5, name='x')

    f1 = m.set_objective((x[1]-1)**2 + (x[1] - x[2])**2, name='f1', sense=so.MIN)
    f2 = m.append_objective((x[1]-x[2])**2 + (x[2] - 3)**2, name='f2', sense=so.MIN)

    m.solve(verbose=True, options={'with': 'blackbox', 'obj': (f1, f2), 'logfreq': 50})

    print('f1', f1.get_value())
    print('f2', f2.get_value())

    if sols:
        return dict(solutions=cas_conn.CASTable('allsols').to_frame(),
                    x=x, f1=f1, f2=f2)
    else:
        return f1.get_value()



import sasoptpy as so
import random


def test(cas_conn, **kwargs):
    # Data generation
    n = 80
    p = 0.02

    random.seed(1)

    ARCS = {}
    for i in range(0, n):
        for j in range(0, n):
            if random.random() < p:
                ARCS[i, j] = random.random()

    max_length = 10

    # Model
    model = so.Model("kidney_exchange", session=cas_conn)

    # Sets
    NODES = set().union(*ARCS.keys())
    MATCHINGS = range(1, int(len(NODES)/2)+1)

    # Variables
    UseNode = model.add_variables(NODES, MATCHINGS, vartype=so.BIN,
                                  name="usenode")
    UseArc = model.add_variables(ARCS, MATCHINGS, vartype=so.BIN,
                                 name="usearc")
    Slack = model.add_variables(NODES, vartype=so.BIN, name="slack")

    print('Setting objective...')

    # Objective
    model.set_objective(so.expr_sum((ARCS[i, j] * UseArc[i, j, m]
                                      for [i, j] in ARCS for m in MATCHINGS)),
                        name="total_weight", sense=so.MAX)

    print('Adding constraints...')
    # Constraints
    Node_Packing = model.add_constraints((UseNode.sum(i, '*') + Slack[i] == 1
                                          for i in NODES), name="node_packing")
    Donate = model.add_constraints((UseArc.sum(i, '*', m) == UseNode[i, m]
                                    for i in NODES
                                    for m in MATCHINGS), name="donate")
    Receive = model.add_constraints((UseArc.sum('*', j, m) == UseNode[j, m]
                                     for j in NODES
                                     for m in MATCHINGS), name="receive")
    Cardinality = model.add_constraints((UseArc.sum('*', '*', m) <= max_length
                                         for m in MATCHINGS),
                                        name="cardinality")

    # Solve
    model.solve(options={'with': 'milp', 'maxtime': 300}, **kwargs)

    # Define decomposition blocks
    for i in NODES:
        for m in MATCHINGS:
            Donate[i, m].set_block(m-1)
            Receive[i, m].set_block(m-1)
    for m in MATCHINGS:
        Cardinality[m].set_block(m-1)

    model.solve(options={
        'with': 'milp', 'maxtime': 300, 'presolver': 'basic', 
        'decomp': {'method': 'user'}}, **kwargs)

    return model.get_objective_value()



import sasoptpy as so
import pandas as pd


def test(cas_conn):

    # Problem data
    OILS = ['veg1', 'veg2', 'oil1', 'oil2', 'oil3']
    PERIODS = range(1, 7)
    cost_data = [
        [110, 120, 130, 110, 115],
        [130, 130, 110, 90, 115],
        [110, 140, 130, 100,  95],
        [120, 110, 120, 120, 125],
        [100, 120, 150, 110, 105],
        [90, 100, 140,  80, 135]]
    cost = pd.DataFrame(cost_data, columns=OILS, index=PERIODS).transpose()
    hardness_data = [8.8, 6.1, 2.0, 4.2, 5.0]
    hardness = {OILS[i]: hardness_data[i] for i in range(len(OILS))}

    revenue_per_ton = 150
    veg_ub = 200
    nonveg_ub = 250
    store_ub = 1000
    storage_cost_per_ton = 5
    hardness_lb = 3
    hardness_ub = 6
    init_storage = 500

    # Problem initialization
    m = so.Model(name='food_manufacture_1', session=cas_conn)

    # Problem definition
    buy = m.add_variables(OILS, PERIODS, lb=0, name='buy')
    use = m.add_variables(OILS, PERIODS, lb=0, name='use')
    manufacture = m.add_implicit_variable((use.sum('*', p) for p in PERIODS),
                                          name='manufacture')
    last_period = len(PERIODS)
    store = m.add_variables(OILS, [0] + list(PERIODS), lb=0, ub=store_ub,
                            name='store')
    for oil in OILS:
        store[oil, 0].set_bounds(lb=init_storage, ub=init_storage)
        store[oil, last_period].set_bounds(lb=init_storage, ub=init_storage)
    VEG = [i for i in OILS if 'veg' in i]
    NONVEG = [i for i in OILS if i not in VEG]
    revenue = so.expr_sum(revenue_per_ton * manufacture[p] for p in PERIODS)
    rawcost = so.expr_sum(cost.at[o, p] * buy[o, p]
                           for o in OILS for p in PERIODS)
    storagecost = so.expr_sum(storage_cost_per_ton * store[o, p]
                               for o in OILS for p in PERIODS)
    m.set_objective(revenue - rawcost - storagecost, sense=so.MAX,
                    name='profit')

    # Constraints
    m.add_constraints((use.sum(VEG, p) <= veg_ub for p in PERIODS),
                      name='veg_ub')
    m.add_constraints((use.sum(NONVEG, p) <= nonveg_ub for p in PERIODS),
                      name='nonveg_ub')
    m.add_constraints((store[o, p-1] + buy[o, p] == use[o, p] + store[o, p]
                      for o in OILS for p in PERIODS),
                      name='flow_balance')
    m.add_constraints((so.expr_sum(hardness[o]*use[o, p] for o in OILS) >=
                      hardness_lb * manufacture[p] for p in PERIODS),
                      name='hardness_ub')
    m.add_constraints((so.expr_sum(hardness[o]*use[o, p] for o in OILS) <=
                      hardness_ub * manufacture[p] for p in PERIODS),
                      name='hardness_lb')

    # Solver call
    res = m.solve()

    # With other solve options
    m.solve(options={'with': 'lp', 'algorithm': 'PS'})
    m.solve(options={'with': 'lp', 'algorithm': 'IP'})
    m.solve(options={'with': 'lp', 'algorithm': 'NS'})

    if res is not None:
        print(so.get_solution_table(buy, use, store))

    return m.get_objective_value()



import sasoptpy as so
import pandas as pd
import math


def test(cas_conn):
    # Input data
    demand_data = pd.DataFrame([
        [0, 2000, 1500, 1000],
        [1, 1000, 1400, 1000],
        [2, 500, 2000, 1500],
        [3, 0, 2500, 2000]
        ], columns=['period', 'unskilled', 'semiskilled', 'skilled'])\
        .set_index(['period'])
    worker_data = pd.DataFrame([
        ['unskilled',   0.25, 0.10, 500, 200, 1500, 50, 500],
        ['semiskilled', 0.20, 0.05, 800, 500, 2000, 50, 400],
        ['skilled',     0.10, 0.05, 500, 500, 3000, 50, 400]
        ], columns=['worker', 'waste_new', 'waste_old', 'recruit_ub',
                    'redundancy_cost', 'overmanning_cost', 'shorttime_ub',
                    'shorttime_cost']).set_index(['worker'])
    retrain_data = pd.DataFrame([
        ['unskilled', 'semiskilled', 200, 400],
        ['semiskilled', 'skilled', math.inf, 500],
        ], columns=['worker1', 'worker2', 'retrain_ub', 'retrain_cost']).\
        set_index(['worker1', 'worker2'])
    downgrade_data = pd.DataFrame([
        ['semiskilled', 'unskilled'],
        ['skilled', 'semiskilled'],
        ['skilled', 'unskilled']
        ], columns=['worker1', 'worker2'])

    semiskill_retrain_frac_ub = 0.25
    downgrade_leave_frac = 0.5
    overmanning_ub = 150
    shorttime_frac = 0.5

    # Sets
    WORKERS = worker_data.index.tolist()
    PERIODS0 = demand_data.index.tolist()
    PERIODS = PERIODS0[1:]
    RETRAIN_PAIRS = [i for i, _ in retrain_data.iterrows()]
    DOWNGRADE_PAIRS = [(row['worker1'], row['worker2'])
                       for _, row in downgrade_data.iterrows()]

    waste_old = worker_data['waste_old']
    waste_new = worker_data['waste_new']
    redundancy_cost = worker_data['redundancy_cost']
    overmanning_cost = worker_data['overmanning_cost']
    shorttime_cost = worker_data['shorttime_cost']
    retrain_cost = retrain_data['retrain_cost'].unstack(level=-1)

    # Initialization
    m = so.Model(name='manpower_planning', session=cas_conn)

    # Variables
    numWorkers = m.add_variables(WORKERS, PERIODS0, name='numWorkers', lb=0)
    demand0 = demand_data.loc[0]
    for w in WORKERS:
        numWorkers[w, 0].set_bounds(lb=demand0[w], ub=demand0[w])
    numRecruits = m.add_variables(WORKERS, PERIODS, name='numRecruits', lb=0)
    worker_ub = worker_data['recruit_ub']
    for w in WORKERS:
        for p in PERIODS:
            numRecruits[w, p].set_bounds(ub=worker_ub[w])
    numRedundant = m.add_variables(WORKERS, PERIODS, name='numRedundant', lb=0)
    numShortTime = m.add_variables(WORKERS, PERIODS, name='numShortTime', lb=0)
    shorttime_ub = worker_data['shorttime_ub']
    for w in WORKERS:
        for p in PERIODS:
            numShortTime.set_bounds(ub=shorttime_ub[w])
    numExcess = m.add_variables(WORKERS, PERIODS, name='numExcess', lb=0)

    retrain_ub = pd.DataFrame()
    for i in PERIODS:
        retrain_ub[i] = retrain_data['retrain_ub']
    numRetrain = m.add_variables(RETRAIN_PAIRS, PERIODS, name='numRetrain',
                                 lb=0, ub=retrain_ub)

    numDowngrade = m.add_variables(DOWNGRADE_PAIRS, PERIODS,
                                   name='numDowngrade', lb=0)
    # Constraints
    m.add_constraints((numWorkers[w, p]
                      - (1-shorttime_frac) * numShortTime[w, p]
                      - numExcess[w, p] == demand_data.loc[p, w]
                      for w in WORKERS for p in PERIODS), name='demand')
    m.add_constraints((numWorkers[w, p] ==
                      (1 - waste_old[w]) * numWorkers[w, p-1]
                      + (1 - waste_new[w]) * numRecruits[w, p]
                      + (1 - waste_old[w]) * numRetrain.sum('*', w, p)
                      + (1 - downgrade_leave_frac) *
                      numDowngrade.sum('*', w, p)
                      - numRetrain.sum(w, '*', p)
                      - numDowngrade.sum(w, '*', p)
                      - numRedundant[w, p]
                      for w in WORKERS for p in PERIODS),
                      name='flow_balance')
    m.add_constraints((numRetrain['semiskilled', 'skilled', p] <=
                      semiskill_retrain_frac_ub * numWorkers['skilled', p]
                      for p in PERIODS), name='semiskill_retrain')
    m.add_constraints((numExcess.sum('*', p) <= overmanning_ub
                      for p in PERIODS), name='overmanning')
    # Objectives
    redundancy = so.Expression(numRedundant.sum('*', '*'), name='redundancy')
    cost = so.Expression(so.expr_sum(redundancy_cost[w] * numRedundant[w, p] +
                                      shorttime_cost[w] * numShortTime[w, p] +
                                      overmanning_cost[w] * numExcess[w, p]
                                      for w in WORKERS for p in PERIODS)
                         + so.expr_sum(
                             retrain_cost.loc[i, j] * numRetrain[i, j, p]
                             for i, j in RETRAIN_PAIRS for p in PERIODS),
                         name='cost')

    m.set_objective(redundancy, sense=so.MIN, name='redundancy_obj')
    res = m.solve()
    if res is not None:
        print('Redundancy:', redundancy.get_value())
        print('Cost:', cost.get_value())
        print(so.get_solution_table(
            numWorkers, numRecruits, numRedundant, numShortTime, numExcess))
        print(so.get_solution_table(numRetrain))
        print(so.get_solution_table(numDowngrade))

    m.set_objective(cost, sense=so.MIN, name='cost_obj')
    res = m.solve()
    if res is not None:
        print('Redundancy:', redundancy.get_value())
        print('Cost:', cost.get_value())
        print(so.get_solution_table(numWorkers, numRecruits, numRedundant,
                                    numShortTime, numExcess))
        print(so.get_solution_table(numRetrain))
        print(so.get_solution_table(numDowngrade))

    return m.get_objective_value()
